



function RosettePDFWriter(rosette, drawGrid, gridWidth, gridColor, units)
{
    var self = this;
    
    self.rosette = rosette;
    
    var displayScale = 72;       // pixels / inch
    var marginX = 0.5 * displayScale;
    var marginY = 0.8 * displayScale;
    
    
    self.getPDFBlob = async function(rosetteFileName)
    {
        
        var doc = new PDFDocument({margin : 0, autoFirstPage: false});

        doc.on('pageAdded', function() {doc.moveDown(); doc.moveDown(); doc.text(rosetteFileName, {align: 'center'});});
        
        // pipe the document to a blob
        var stream = doc.pipe(blobStream());
        
        // stupid hack to get stream result through promise...
        var done = new Promise(function(resolve, reject) 
        {
            stream.on('finish', resolve);
            stream.on('error', reject);
        });
        
        // generate pages
        //self.generateTileStrip(doc, self.rosette);
        
        doc.addPage();
        self.generateRosette(doc, self.rosette);

        doc.addPage();
        self.generateTileMap(doc, self.rosette);
        
        doc.addPage();
        self.generateSpecifications(doc, self.rosette);
        
        // get a blob when you're done
        doc.end();
        
        // and now the asynchronous hack - wait on Promise, which will be fulfilled when stream ends... gah...
        var conclusion = await done;
        
        var blob =  stream.toBlob('application/pdf');
        
        return blob;
        
    }
    
    
    self.generateTileStrip = function(doc, rosette)
    {
        
    }
    
    self.generateRosette = function(doc, rosette)
    {
        var pi = Math.PI;
        var numTiles;
        var numRows, numCols;
        var tileHeight, tileWidth;
        var scaledRowHeight, scaledTileHeight;
        var rIn, rOut, theta;
        var u, v, w, v1, v2, w1, w2;
        var elementColor;
        
        
        numRows = rosette.getRows();
        numCols = rosette.getCols();

        tileHeight = rosette.getTileHeight();
        tileWidth = rosette.getTileWidth();

        scaledTileHeight = tileHeight * displayScale;
        scaledRowHeight = scaledTileHeight / numRows;
        // height of tile and row, in fractional pixels

        rIn = rosette.radius * displayScale;
        rOut = rosette.getOuterRadius() * displayScale;
        // outer rosette radius, in fractional pixels
        
        
        // translate to page center
        doc.translate(doc.page.width/2, rOut + marginY);
        
        
        u = new RealPoint(0, -rOut);
        // vector to outer radius at 12 o'clock

        numTiles = Math.floor(pi * rOut / (tileWidth * displayScale));
        // number of tiles which fit into half of rosette, rounded down

        theta = (tileWidth * displayScale / numCols) / rOut;
        // angle subtended by one column-width of tile

        for(var i = -numTiles; i < numTiles; ++i)
        {
            w = u.rotate(i*numCols*theta);
            for (var column = 0; column < numCols; ++column)
            {
                v = w.duplicate();
                w = v.rotate(theta);
                // v2 = v.duplicate();
                // w2 = w.duplicate();
                for (var row = 0; row < numRows; ++ row)
                {
                    if (i % 2 == 0) // even tile
                        elementColor = rosette.getEvenElementColor(row,column);
                    else            // odd tile
                        elementColor = rosette.getOddElementColor(row,column);
                    
                    
                    v1 = v.scalarMult(1 - row * scaledRowHeight / rOut);
                    v2 = v.scalarMult(1 - (row+1) * scaledRowHeight / rOut);
                    w1 = w.scalarMult(1 - row * scaledRowHeight / rOut);
                    w2 = w.scalarMult(1 - (row+1) * scaledRowHeight / rOut);
                    var xCoords = [v1.x, w1.x, w2.x, v2.x];
                    var yCoords = [-v1.y, -w1.y, -w2.y, -v2.y];

                    
                    doc.polygon([v1.x, v1.y], [w1.x, w1.y], [w2.x, w2.y], [v2.x, v2.y]).fill(elementColor.color.toHexString());
                    
                    if (false /*drawGrid*/)
                    {
                        doc.polygon([v1.x, v1.y], [w1.x, w1.y], [w2.x, w2.y], [v2.x, v2.y])
                            .lineWidth(gridWidth*displayScale)
                            .stroke(gridColor.color.toHexString());
                    }
                    
                }
            }
        }
        
        
        // draw inner and outer rings
        var innerRings = self.rosette.innerRings;
        var innerRadius = rOut-scaledTileHeight;
        for (var i = 0; i < innerRings.length; i++)
        {
            // go from inner radius inward
            var ring = innerRings[i];
            var strokeWidth = ring.width * displayScale;
            var strokeColor = ring.namedColor.color.toHexString();
            
            // stroke width is centered along path, so offset by half stroke width
            innerRadius -= (ring.width/2 * displayScale);
            doc.circle(0, 0, innerRadius).lineWidth(strokeWidth).stroke(strokeColor);

            innerRadius -= (ring.width/2 * displayScale);
        }
        
        var outerRings = self.rosette.outerRings;
        var outerRadius = rOut;
        for (var i = 0; i < outerRings.length; i++)
        {
            // go from outer radius inward
            var ring = outerRings[i];
            var strokeWidth = ring.width * displayScale;
            var strokeColor = ring.namedColor.color.toHexString();
            
            // stroke width is centered along path, so offset by half stroke width
            outerRadius += (ring.width/2 * displayScale);
            doc.circle(0, 0, outerRadius).lineWidth(strokeWidth).stroke(strokeColor);

            outerRadius += (ring.width/2 * displayScale);
        }
        
        
        // outline the rosette; helps when there are no rings
        doc.lineWidth(1);
        doc.circle(0, 0, innerRadius).stroke("black");
        doc.circle(0, 0, outerRadius).stroke("black");
        
        
        // draw an "x" at the center point
        var lineLength = 0.1 * displayScale;
        doc.moveTo(-lineLength, 0).lineTo(lineLength, 0).stroke("black");
        doc.moveTo(0, -lineLength).lineTo(0, lineLength).stroke("black");

    }
    
    
    
    
    self.generateTileMap = function(doc, rosette)
    {
     
        
        var numRows = rosette.getRows();
        var numCols = rosette.getCols();
        var colors = new Array();
        var counts = new Array();
        
        
        // set origin at the edge of the margin and header
        doc.translate(marginX, marginY);
        
        
        // make a list of the colors used
        for (var i = 0; i < numRows; i++)
        {
            for (var j = 0; j < numCols; j++)
            {
                var nextColor = rosette.getEvenElementColor(i, j);
                var index = colors.indexOf(nextColor);
                if (index == -1)
                {
                    colors.push(nextColor);
                }
            }
        }
        
        
        // get page params
        var pageImageableWidth = (doc.page.width - 2 * marginX) / 72;
        var pageImageableHeight = (doc.page.height - 2 * marginY) / 72;
        //double pageImageableX = pageFormat.getImageableX() / 72;
        //double pageImageableY = pageFormat.getImageableY() / 72;
        
        //double scaleFactor = 5;
        
        // get Graphics2d object to scale image
        //Graphics2D g2d = (Graphics2D)g;
        //g2d.scale(1/scaleFactor, 1/scaleFactor);
        
        //int displayScale = (int)Math.round(72 * scaleFactor);       // pixels / inch
        
        
        // determine sizes of squares
        var squareWidth = Math.floor(displayScale * pageImageableWidth/numCols * 3/5);
        if (squareWidth > Math.floor(displayScale * pageImageableHeight/numRows * 3/5))
            squareWidth = Math.floor(displayScale * pageImageableHeight/numRows * 3/5);
        
        // determine sizes of gaps between squares
        var squareHorizontalGap = Math.floor(displayScale * pageImageableWidth/numCols * 2/5);
        var squareVerticalGap = Math.floor(displayScale * pageImageableHeight/numRows * 2/5);
        
        
        var x = 0;
        var y = 0;
        
        // set font size to 8 point
        //g.setFont(g.getFont().deriveFont((float)(8.0 * scaleFactor)));
        
        //FontMetrics fontMetrics = g.getFontMetrics();
        //int fontHeight = fontMetrics.getAscent();
        //int fontBaseline = (fontHeight + squareWidth)/2;
        
            
        for (var i = 0; i < numRows; i++)
        {
            for (var j = 0; j < numCols; j++)
            {
                var elementColor = rosette.getElementColor(i, j);
                
                doc.rect(x, y, squareWidth, squareWidth).fillAndStroke(elementColor.color.toHexString(), "black");
                
                /*
                var textColor;
                if ((elementColor.getRed() + elementColor.getGreen() + elementColor.getBlue()) > (255) )
                {
                    textColor = Color.black;
                }
                else
                {
                    textColor = Color.white;
                }
                
                if (printParameters.printTileMapNumbers)
                {
                    int colorIndex = colors.indexOf(elementColor) + 1;
                    String colorIndexString = Integer.toString(colorIndex);
                    
                    int stringWidth = fontMetrics.stringWidth(colorIndexString);
                    int stringStart = (squareWidth - stringWidth)/2;
                    
                    g.setColor(textColor);
                    g.drawString(Integer.toString(colorIndex), x + stringStart, y + fontBaseline);
                    g.setColor(Color.black);
                }
                */
                
                x += squareWidth + squareHorizontalGap;
                
            }
            
            x = 0;
            y += squareWidth + squareVerticalGap;
        }
        
        //g2d.scale(scaleFactor, scaleFactor);
    }
    
      
    
    self.generateSpecifications = function(doc, rosette)
    {
        var numRows = rosette.getRows();
        var numCols = rosette.getCols();
        var colors = new Array();
        var counts = new Array();
        
        
        // set up number format
        //NumberFormat numberFormat = NumberFormat.getInstance();
        //numberFormat.setMaximumFractionDigits(3);
        //numberFormat.setMinimumFractionDigits(3);

        
        // set origin at the edge of the margin and header
        //var marginX = 50;
        //var marginY = 50;
        
        doc.translate(marginX, marginY);
        
        // put title on top of page
        //doc.fontSize(12);
        //doc.text(rosetteFileName);
        
        
        // make a list of the colors used
        for (var i = 0; i < numRows; i++)
        {
            for (var j = 0; j < numCols; j++)
            {
                var nextColor = rosette.getEvenElementColor(i, j);
                
                var index = colors.indexOf(nextColor);
                
                if (index == -1)
                {
                    colors.push(nextColor);
                }
                
            }
        }
        
        // now get counts of colors used
        counts = new Array(colors.length);
        for (var i = 0; i < counts.length; i++)
        {
            counts[i] = 0;
        }
        
        for (var i = 0; i < numRows; i++)
        {
            for (var j = 0; j < numCols; j++)
            {
                var nextColor = rosette.getEvenElementColor(i, j);
                
                var index = colors.indexOf(nextColor);
                
                counts[index]++;
            }
        }
        
        
        // get maximum width of color name strings
        var maxStringWidth = 0;
        for (var i = 0; i < colors.length; i++)
        {
            
            var name = colors[i].name;
            var stringWidth = doc.widthOfString(name);
            if (stringWidth > maxStringWidth)
                maxStringWidth = stringWidth;
        }
        
        
        // determine sizes of squares - make them 1/2"
        var squareWidth = Math.floor(displayScale * 0.5);
        
        // draw squares
        var x = 0;
        var y = squareWidth/2;
        
        doc.font('Helvetica-Bold');
        doc.text("Color Counts", x, y);
        doc.font('Helvetica');
        
        
        y += squareWidth/2;
        
        
        for (var i = 0; i < counts.length; i++)
        {
            var elementColor = colors[i].color;
            
            // print a square with the color, followed by the count
            doc.rect(x, y, squareWidth, squareWidth).fillAndStroke(elementColor.toHexString(), "black");
            
            y += squareWidth/2;
            

            // color back to black; note that this is case-sensitive... :-(
            doc.fillColor("black").strokeColor("black");
            
            var colorNameString = colors[i].name;
            doc.text(colorNameString, x + squareWidth + squareWidth, y);
            
            var colorCountString = counts[i];
            doc.text(colorCountString, x + squareWidth + squareWidth + squareWidth + maxStringWidth, y);
            
            //System.out.println("Color count " + i + ": " + counts[i]);
            
            y += squareWidth;// + squareWidth/2;
        }
        
        y += squareWidth/2;
        
        // print dimensions
        var radius = rosette.getRadius();
        
        // tile height, tile width top and bottom
        var tileHeight = rosette.getTileHeight();
        var tileWidthTop = rosette.getTileWidth();
        var tileWidthBottom = tileWidthTop * (rosette.getRadius() / (rosette.getRadius() + rosette.getTileHeight()));
        
        // row height, column thicknesses at top and bottom
        var rowHeight = tileHeight/rosette.getRows();
        var columnWidthTop = tileWidthTop/rosette.getCols();
        var columnWidthBottom = columnWidthTop * (rosette.getRadius() / (rosette.getRadius() + rosette.getTileHeight()));
        
        var unitsString = " " + units.toUnitsString();
        var decimalPlaces = units.decimalPlaces();
        
        // move back to start of page, below the squares
        //doc.translate(-x + squareWidth + squareWidth + squareWidth + maxStringWidth, 0);
        //doc.restore();
        
        doc.fontSize(12);
        var delta = 20;
        var offset = 150;
        
        // print parameters in desired units
        doc.font('Helvetica-Bold');
        doc.text("Dimensions", 0, y);
        
        doc.font('Helvetica');
        
        y += delta;
        doc.text("Number of rows:", 0, y);
        doc.text(numRows, offset, y);
        
        y += delta;
        doc.text("Number of columns:", 0, y);
        doc.text(numCols, offset, y);
        
        y += delta;
        doc.text("Radius:", 0, y);
        doc.text(units.toUnits(radius).toFixed(decimalPlaces) + unitsString, offset, y);
        
        y += delta;
        doc.text("Tile height:", 0, y);
        doc.text(units.toUnits(tileHeight).toFixed(decimalPlaces) + unitsString, offset, y);
        
        y += delta;
        doc.text("Tile width at top:", 0, y);
        doc.text(units.toUnits(tileWidthTop).toFixed(decimalPlaces) + unitsString, offset, y);
        
        y += delta;
        doc.text("Tile width at bottom:", 0, y);
        doc.text(units.toUnits(tileWidthBottom).toFixed(decimalPlaces) + unitsString, offset, y);
        
        y += delta;
        doc.text("Row height:", 0, y);
        doc.text(units.toUnits(rowHeight).toFixed(decimalPlaces) + unitsString, offset, y);
        
        y += delta;
        doc.text("Column width at top:", 0, y);
        doc.text(units.toUnits(columnWidthTop).toFixed(decimalPlaces) + unitsString, offset, y);
        
        y += delta;
        doc.text("Column width at bottom:", 0, y);
        doc.text(units.toUnits(columnWidthBottom).toFixed(decimalPlaces) + unitsString, offset, y);
    }
    
}